#include "entropycalculator.h"
#include "streamlinepool.h"
#include "unitspherepartitioner.h"
#include <unordered_map>
#include <string>
#include <algorithm>
#include <cassert>

EntropyCalculator& EntropyCalculator::instance()
{
	static EntropyCalculator calculator;
	return calculator;
}

void EntropyCalculator::setStreamlinePool(shared_ptr<StreamlinePool> pool) 
{ 
	_pStreamlines = pool;
	for (auto s : pool->streamlines) {
		assert(s->attributes.find("velocityBin") == s->attributes.end());
		Attribute<unsigned>* bin_info = new Attribute<unsigned>(s->numPts);
		Attribute<float3>* velocities = dynamic_cast<Attribute<float3>*>(s->attributes["velocities"]);

		for (size_t i = 0; i < velocities->size(); i++)
			(*bin_info)[i] = UnitSpherePartitioner::instance().findRegionForVector((*velocities)[i]);


		s->attributes["velocityBin"] = bin_info;
	}
}

void EntropyCalculator::computeEntropyRatio(std::shared_ptr<StreamlinePool> pool, unsigned nhs)
{
	
	std::string propName = "velocityEntropyRatio" + std::to_string(nhs);
	
	for (auto s : pool->streamlines) {
		assert(s->attributes.find("simplification_points") != s->attributes.end());
		assert(s->attributes.find(propName) == s->attributes.end());
		Attribute<unsigned>* indices = dynamic_cast<Attribute<unsigned>*>(s->attributes["simplification_points"]);
		Attribute<float>* entropy = new Attribute<float>(indices->size());
		
		for (auto i = 0; i < indices->size(); i++) {
			float entropy1 = computeEntropy_helper(s, std::max(unsigned(0), (*indices)[i] - nhs), (*indices)[i]);
			float entropy2 = computeEntropy_helper(s, (*indices)[i], std::min(s->numPts - 1, (*indices)[i] + nhs));
			assert(entropy1 == 0 && entropy2 == 0);
			if (entropy1 > entropy2)
				std::swap(entropy1, entropy2);
			(*entropy)[i] = entropy1 / entropy2;
		}
		
		s->attributes[propName] = entropy;
	}
	
}

float EntropyCalculator::computeEntropyRatio(int sid, int center, int nhs)
{
	Streamline* s = _pStreamlines->streamlines[sid];
	if (center - nhs < 0)
		nhs = center;
	if (center + nhs > int(s->numPts) - 1)
		nhs = int(s->numPts) - 1 - center;
	float entropy1 = computeEntropy_helper(s, center-nhs, center);
	float entropy2 = computeEntropy_helper(s, center, center+nhs);
	if (entropy1 > entropy2)
		std::swap(entropy1, entropy2);

	if (entropy1 == 0 && entropy2 == 0)
		return 1.0;
	else
		return entropy1 / entropy2;
}

float EntropyCalculator::computeEntropy_helper(Streamline* stl, unsigned start, unsigned end) {

	Attribute<unsigned>* bin_info = dynamic_cast<Attribute<unsigned>*>(stl->attributes["velocityBin"]);
	std::unordered_map<unsigned, unsigned> distribution;


	for (unsigned i = start; i <= end; i++) {
		int velReg = (*bin_info)[i];
		if (distribution.find(velReg) == distribution.end())
			distribution[velReg] = 1;
		else
			distribution[velReg]++;
	}

	float entropy = 0;
	for (auto it = distribution.begin(); it != distribution.end(); ++it) {
		float p = float(it->second) / float(end - start + 1);
		entropy += p*log(p);
	}

	return -entropy;
}