#include "minvolenclosingellipsoid.h"
#include "streamlinepool.h"
#include <Eigen/LU>
#include <Eigen/SVD>
#include <algorithm>
#include <iostream>


float MVEE::computeBoundingEllipsoidVolRatio(int sid, int center, int nhs)
{
	if (nhs < 2)
		return 1.0;

	vector<float3> points = dynamic_cast<Attribute<float3>*>(_pStreamlines->streamlines[sid]->attributes["points"])->data();
	float totalVol = computeEllipsoidVolume(points, center - nhs, center + nhs);
	if (totalVol == 0)
		return 1.0;
	else {
		float vol1 = computeEllipsoidVolume(points, center - nhs, center);
		float vol2 = computeEllipsoidVolume(points, center, center + nhs);
		return (vol1 + vol2) / totalVol;
	}
}

float MVEE::computeBoundingEllipsoidVolRatio(int sid, int start, int mid, int end) {
	vector<float3> points = dynamic_cast<Attribute<float3>*>(_pStreamlines->streamlines[sid]->attributes["points"])->data();
	float totalVol = computeEllipsoidVolume(points, start, end);
	if (totalVol == 0)
		return 1.0;
	else {
		float vol1 = computeEllipsoidVolume(points, start, mid);
		float vol2 = computeEllipsoidVolume(points, mid, end);
		return (vol1 + vol2) / totalVol;
	}
}

float MVEE::computeEllipsoidVolume(const vector<float3>& points, int start, int end)
{
	int n = end - start + 1;
	Eigen::Matrix3Xf P(3, n);
	for (int i = 0; i < n; i++) {
		int id = start + i;
		float3 pt = points[id];
		P(0, i) = pt.x;
		P(1, i) = pt.y;
		P(2, i) = pt.z;
	}

	Ellipsoid ellipsoid;
	bool non_degenerate = boundingEllipsoid(P, ellipsoid.Q, ellipsoid.c);
	if (non_degenerate) {
		computeEllipsoidParameters(ellipsoid);
		float ratio = ellipsoid.radii(2) / ellipsoid.radii(0);
		if (ratio <= 0.03)
			return 0;
		else
			return ellipsoidVolume(ellipsoid.radii);
	}
	else
		return 0;
}


bool MVEE::boundingEllipsoid(const Eigen::Matrix3Xf &P, Eigen::Matrix3f &A, Eigen::Vector3f &c, float tolerance)
{
	int N = P.cols();   // total number of points
	int d = 3;

	// lifting
	Eigen::Matrix4Xf Q(4, N);
	Q.topLeftCorner(3, N) = P;
	Q.row(3) = Eigen::VectorXf::Constant(N, 1.0);

	// initialize
	float err = 1.0;
	Eigen::VectorXf u = Eigen::VectorXf::Constant(N, 1.0 / N);

	while (err > tolerance) {
		Eigen::MatrixX4f Q_trans = Q.transpose();
		Eigen::Matrix4f X = Q * u.asDiagonal() * Q_trans;

		Eigen::Matrix4f X_inv;
		float determinant;
		bool invertible;
		X.computeInverseAndDetWithCheck(X_inv, determinant, invertible);

		if (!invertible)
			return false;

		Eigen::VectorXf m = (Q_trans * X_inv * Q).diagonal();
		int j = std::max_element(m.data(), m.data() + N) - m.data();
		float maximum = m(j);

		float step_size = (maximum - d - 1) / ((d + 1) * (maximum - 1));
		Eigen::VectorXf new_u = (1 - step_size) * u;
		new_u(j) += step_size;

		err = (new_u - u).norm();
		u = new_u;
	}

	Eigen::Matrix3f tmp = P * u.asDiagonal() * P.transpose() - (P * u) * (P * u).transpose();
	Eigen::Matrix3f tmp_inv;
	float determinant;
	bool invertible;
	tmp.computeInverseAndDetWithCheck(tmp_inv, determinant, invertible);
	if (!invertible)
		return false;

	A = (1.0 / d) * tmp_inv;
	c = P * u;

	return true;

}


void MVEE::computeEllipsoidParameters(Ellipsoid & ellipsoid)
{
	Eigen::JacobiSVD<Eigen::Matrix3f> svd(ellipsoid.Q, Eigen::ComputeFullU | Eigen::ComputeFullV);
	ellipsoid.axes.col(0) = svd.matrixV() * Eigen::Vector3f::UnitZ();
	ellipsoid.axes.col(1) = svd.matrixV() * Eigen::Vector3f::UnitY();
	ellipsoid.axes.col(2) = svd.matrixV() * Eigen::Vector3f::UnitX();
	ellipsoid.radii(0) = 1.0 / sqrt((svd.singularValues())(2));
	ellipsoid.radii(1) = 1.0 / sqrt((svd.singularValues())(1));
	ellipsoid.radii(2) = 1.0 / sqrt((svd.singularValues())(0));

}


float MVEE::ellipsoidVolume(const Eigen::Vector3f& radii) { return 4.0 * 3.1415926 * radii(0) * radii(1) * radii(2) / 3.0; }




