#include "streamlinepool.h"
#include "curvtorscalculator.h"

#include <fstream>
#include <string>
#include <ctime>
#include <cstdlib>
#include <iostream>
#include <algorithm>



const unsigned MAX_POINT_NUM = 500;
const unsigned MIN_POINT_NUM = 20;
const float    TRACE_INTERVAL = 0.05f;
const float    SEG_LEN = 0.5f;
const unsigned MAX_STEP_NUM = 5000;


StreamlinePool::~StreamlinePool()
{
    for(auto pStl : streamlines)
        delete pStl;
}

bool StreamlinePool::readVelocityField(const char* filename)
{
    fstream input;

    // find and read the header file first
    string vfFile(filename);
    string hdrFile = vfFile.replace(vfFile.size()-3,3,"hdr");

    input.open(hdrFile.c_str());
    if (input.is_open()) {
        input >> width >> height >> depth;
        input.close();
    }
    else
        return false;

    // read velocity field
    input.open(filename, ios::in | ios::binary);
    if (input.is_open()) {
        int total = width * height * depth;
        _velocityField = move(unique_ptr<float3[]>(new float3[total]));
        input.read((char*)_velocityField.get(), total*sizeof(float3));
		return true;
    }
    else
        return false;

}

void StreamlinePool::trace(unsigned numStl, bool useMagicNumber)
{
    // generate random seeds
    if( !useMagicNumber )
        magicNumber = (unsigned int)time(nullptr);

    
    srand(magicNumber);

    unsigned count = 0;
    float3 seed;
    while(count<numStl){
           seed.x = rand()%(100*(width-1))/100.0f;
           seed.y = rand()%(100*(height-1))/100.0f;
           seed.z = rand()%(100*(depth-1))/100.0f;
		   if (traceSingleStreamlineFrom(seed)) {
			   streamlines.back()->id = count;
			   count++;
		   }
    }


	center = make_float3((_minx + _maxx) / 2.0, (_miny + _maxy) / 2.0, (_minz + _maxz) / 2.0);
	

}


bool StreamlinePool::isOutOfBound(const float3 &pos){
   if (pos.x<=0.5000001f || pos.x>=(width-1.5000001f) ||
       pos.y<=0.5000001f || pos.y>=(height-1.5000001f) ||
       pos.z<=0.5000001f || pos.z>=(depth-1.5000001f))
       return true;
   else
    return false;
}

bool StreamlinePool::vecAt(float3 &ret, const float3 &pos){
   if (pos.x<=0.5000001f || pos.x>=(width-1.5000001f) ||
       pos.y<=0.5000001f || pos.y>=(height-1.5000001f) ||
       pos.z<=0.5000001f || pos.z>=(depth-1.5000001f))
       return false;


   int x = int(floorf(pos.x));
   int y = int(floorf(pos.y));
   int z = int(floorf(pos.z));
   int idx1 = x+y*width+z*width*height;
   int idx2 = x+1+y*width+z*width*height;
   int idx3 = x+(y+1)*width+z*width*height;
   int idx4 = x+1+(y+1)*width+z*width*height;
   int idx5 = x+y*width+(z+1)*width*height;
   int idx6 = x+1+y*width+(z+1)*width*height;
   int idx7 = x+(y+1)*width+(z+1)*width*height;
   int idx8 = x+1+(y+1)*width+(z+1)*width*height;

   float facx=pos.x-x, facy=pos.y-y, facz=pos.z-z;

   ret = (1-facx)*(1-facy)*(1-facz)*_velocityField[idx1]\
           +facx*(1-facy)*(1-facz)*_velocityField[idx2]\
           +(1-facx)*facy*(1-facz)*_velocityField[idx3]\
           +facx*facy*(1-facz)*_velocityField[idx4]\
           +(1-facx)*(1-facy)*facz*_velocityField[idx5]\
           +facx*(1-facy)*facz*_velocityField[idx6]\
           +(1-facx)*facy*facz*_velocityField[idx7]\
           +facx*facy*facz*_velocityField[idx8];
   return true;
}

bool StreamlinePool::nextPoint(float3 &ret, float3 &pos, bool isFoward){//RK4
   float3 k1, k2, k3, k4;
   float itv;
   if (isFoward) {
       itv = TRACE_INTERVAL;
   } else {
       itv = -TRACE_INTERVAL;
   }

   if (vecAt(k1, pos)) {
       if (vecAt(k2, pos+0.5f*itv*k1)) {
           if (vecAt(k3, pos+0.5f*itv*k2)) {
               if (vecAt(k4, pos+itv*k3)) {
                   ret = pos+(itv/6.0f)*(k1+2*k2+2*k3+k4);
                   return true;
               }
           }
       }
   }
   return false;
}

bool StreamlinePool::traceSingleStreamlineFrom(float3 &pos){
   float3 parray[MAX_POINT_NUM];
   float3 varray[MAX_POINT_NUM];
   float3 lastPos = pos;
   parray[0] = pos;
   if(!vecAt(varray[0], pos)) return false;

   unsigned count=1, stepCount=0, numBack=0;
   //backward tracing first
   while (nextPoint(pos, pos, false) && count<MAX_POINT_NUM) {
       stepCount++;
       if (length(pos - lastPos)>=SEG_LEN) {
           lastPos = pos;
           parray[count] = pos;
           vecAt(varray[count], pos);
           count++;
           stepCount = 0;
           continue;
       }
       if (stepCount>MAX_STEP_NUM) {
           break;
       }
   }
   numBack = count;
   if(isOutOfBound(parray[numBack-1])){
       numBack--;
       count--;
   }

   //forward tracing
   pos = parray[0];
   lastPos = pos;
   stepCount = 0;
   while (nextPoint(pos, pos, true) && count<MAX_POINT_NUM) {
       stepCount++;
       if (length(pos-lastPos)>=SEG_LEN) {
           lastPos = pos;
           parray[count] = pos;
           vecAt(varray[count], pos);
           count++;
           stepCount = 0;
           continue;
       }
       if (stepCount>MAX_STEP_NUM) {
           break;
       }
   }
   if(isOutOfBound(parray[count-1])){
       count--;
   }

   if (count<MIN_POINT_NUM)
       return false;

   auto points = new Attribute<float3>(count);
   auto velocities = new Attribute<float3>(count);

   //copy backward points
   reverse_copy(parray, parray+numBack, points->begin());
   reverse_copy(varray, varray+numBack, velocities->begin());

   //copy forward points
   copy(parray+numBack, parray+count, points->begin()+numBack);
   copy(varray+numBack, varray+count, velocities->begin()+numBack);

   auto arclength = new Attribute<float>(count);
   float accum_length = .0f;
   (*arclength)[0] = accum_length;
   float3 prev = (*points)[0], next;
   for(size_t i = 1; i < (*points).size(); i++) {
       next = (*points)[i];
       accum_length += length(next-prev);
       prev = next;
       (*arclength)[i] = accum_length;
   }


   Streamline* stl = new Streamline;
   stl->attributes["points"] = points;
   stl->attributes["velocities"] = velocities;
   stl->attributes["arclength"] = arclength;
   streamlines.push_back(stl);

   stl->startIndex = totalNumPoints;
   stl->numPts = count;
   totalNumPoints += count;

   auto cmpX = [](float3& p1, float3& p2) { return p1.x < p2.x; };
   auto cmpY = [](float3& p1, float3& p2) { return p1.y < p2.y; };
   auto cmpZ = [](float3& p1, float3& p2) { return p1.z < p2.z; };

   auto it = std::min_element(points->begin(), points->end(), cmpX);
   _minx = min(_minx, it->x);
   it = std::max_element(points->begin(), points->end(), cmpX);
   _maxx = max(_maxx, it->x);
   
   it = std::min_element(points->begin(), points->end(), cmpY);
   _miny = min(_miny, it->y);
   it = std::max_element(points->begin(), points->end(), cmpY);
   _maxy = max(_maxy, it->y);

   it = std::min_element(points->begin(), points->end(), cmpZ);
   _minz = min(_minz, it->z);
   it = std::max_element(points->begin(), points->end(), cmpZ);
   _maxz = max(_maxz, it->z);

   return true;
}
