#ifndef STREAMLINEPOOL_H
#define STREAMLINEPOOL_H

#include <unordered_map>
#include <vector>
#include <string>
#include <memory>

#include "vector_types.h"

using namespace std;



/*!
  holds scalar/vector values for each point on a streamline
  */
class AttributeBase {
public:
    virtual ~AttributeBase() {}
    virtual size_t size() = 0;
};

template <class T>
class Attribute : public AttributeBase{
public:
    typedef typename vector<T>::const_iterator const_iterator;
    typedef typename vector<T>::iterator iterator;
    typedef typename vector<T>::reverse_iterator reverse_iterator;
public:
	Attribute() {}

    Attribute(unsigned n) {
        _data.resize(n);
    }

    iterator begin() { return _data.begin(); }
    iterator end() { return _data.end(); }
    const_iterator cbegin() { return _data.cbegin(); }
    const_iterator cend() { return _data.cend(); }
    reverse_iterator rbegin() { return _data.rbegin(); }
    reverse_iterator rend() { return _data.rend(); }

    const T& operator[] (unsigned i) const { return _data[i]; }
    T& operator[] (unsigned i) { return _data[i]; }

    size_t size() { return _data.size(); }

	void clear() { _data.clear();  }

	const vector<T>& data() const { return _data;  }
	vector<T>& data() { return _data; }

	void push_back(const T& d) { _data.push_back(d); }

	void pop_back() { _data.pop_back();  }

	const T& front() const { return _data[0]; }
	const T& back() const { return _data.back(); }

private:
    vector<T> _data;
};


class Streamline {
public:
	Streamline() : radius(0.5) {}
    ~Streamline() {
        for(auto it = attributes.begin(); it != attributes.end(); ++it)
            delete it->second;
    }

    unordered_map<string, AttributeBase*> attributes;

	unsigned id;			///< id of this streamline
    unsigned startIndex;    ///< index of the first point in this streamline
    unsigned numPts;        ///< number of points in this streamline
	float	 radius;
};



/*!
  This class holds various information about streamlines
  */
class StreamlinePool {
public:
    StreamlinePool() : magicNumber(123), totalNumPoints(0), _maxx(-10000), _maxy(-10000), _maxz(-10000), _minx(10000), _miny(10000), _minz(10000) {}
    ~StreamlinePool();

    bool readVelocityField(const char* file);
    void trace(unsigned int numStl, bool useMagicNumber);


    vector<Streamline*> streamlines;
    unsigned            width, height, depth;
	float3				center;
    unsigned            magicNumber;
    unsigned            totalNumPoints;

private:
    bool traceSingleStreamlineFrom(float3& pos);
    bool isOutOfBound(const float3& pos);
    bool nextPoint(float3& ret, float3& pos, bool isForward);
    bool vecAt(float3& ret, const float3& pos);

    unique_ptr<float3[]>    _velocityField;

	float					_minx, _miny, _minz, _maxx, _maxy, _maxz;

};



#endif // STREAMLINEPOOL_H


