#include "streamlinesimplifier.h"

#include <unordered_map>
#include <algorithm>


StreamlineSimplifier& StreamlineSimplifier::instance()
{
	static StreamlineSimplifier simplifier;
	return simplifier;
}

void StreamlineSimplifier::simplify(vector<Streamline*>& streamlines, float epsilon)
{
	_epsilon = epsilon;
	for (auto s : streamlines) {
		Attribute<unsigned> *indices;
		if (s->attributes.find("simplification_points") == s->attributes.end()) {
			indices = new Attribute<unsigned>;
			s->attributes["simplification_points"] = indices;
		}
		else {
			indices = dynamic_cast<Attribute<unsigned>*>(s->attributes["simplification_points"]);
			indices->clear();
		}

		Attribute<float3> *points = dynamic_cast<Attribute<float3>*>(s->attributes["points"]);

		indices->push_back(0);
		FS(indices, points->data());

		if (indices->back() != s->numPts - 1) {
			if (s->numPts - indices->back() < 10)
				indices->data().pop_back();
			indices->push_back(s->numPts - 1);
		}

	}
}


void StreamlineSimplifier::FS(Attribute<unsigned> * indices, const vector<float3>& P)
{
	int n = P.size();
	int i = 0, l, j;
	while (i < n) {
		l = 0;
		j = i + int(pow(2.0, l + 1));
		if (j >= n)
			j = n - 1;

		// look for l such that the frechet distance between i and 2^(l+1) is larger than epsilon
		// and the frechet distance between i and 2^l is less than epsilon
		while (frechet_distance_less_than_epsilon(P, i, j)) {
			l += 1;
			if (i + int(pow(2.0, l + 1)) >= n)
				break;
			else
				j = i + int(pow(2.0, l + 1));
		}

		// now try to location a simplification point between i+2^l and i+2^(l+1)
		int low = int(pow(2.0, l));
		int high = int(pow(2.0, l + 1));

		if (i + low >= n - 1)
			break;
		else if (i + high >= n)
			high = n - 1 - i;

		while (low < high - 1) {
			int mid = (low + high) / 2;

			if (frechet_distance_less_than_epsilon(P, i, i + mid))
				low = mid;
			else
				high = mid;
		}
		i += low;

		indices->push_back(i);
	}
}

bool StreamlineSimplifier::frechet_distance_less_than_epsilon(const std::vector<float3>& P, int i, int j)
{
	unordered_map<int, vector<pair<float, float>>> F;
	float d1_x = P[j].x - P[i].x;
	float d1_y = P[j].y - P[i].y;
	float d1_z = P[j].z - P[i].z;

	int counter = 1;
	pair<float, float> B_F;
	float pq_x, pq_y, pq_z;
	for (int k = i + 1; k < j + 1; k++) {
		F[counter].resize(4);

		pq_x = P[i].x - P[k - 1].x;
		pq_y = P[i].y - P[k - 1].y;
		pq_z = P[i].z - P[k - 1].z;

		float a = d1_x * d1_x + d1_y * d1_y + d1_z * d1_z;
		float b = 2 * (pq_x * d1_x + pq_y * d1_y + pq_z * d1_z);
		float c = pq_x * pq_x + pq_y * pq_y + pq_z * pq_z - _epsilon * _epsilon;
		float d = b * b - 4 * a * c;

		if (fabs(d) <= 1e-8) {
			float root = -b / (2.0f * a);
			if (root < 0 || root  > 1)
				return false;
			else
				B_F = std::make_pair(root, root);
		}
		else {
			if (d < 0)
				return false;
			else {
				float root1 = (-b + sqrt(d)) / (2.0f * a);
				float root2 = (-b - sqrt(d)) / (2.0f * a);
				if (root1 > root2)
					std::swap(root1, root2);

				if (root1 > 1 && root2 > 1 || root1 < 0 && root2 < 0)
					return false;
				else {
					if (root1 < 0)
						root1 = 0;
					if (root2 < 1)
						root2 = 1;
					B_F = std::make_pair(root1, root2);
				}
			}
		}

		F[counter][0] = B_F;
		counter++;
	}   //end for loop

	float d2_x = P[j].x - P[j - 1].x;
	float d2_y = P[j].y - P[j - 1].y;
	float d2_z = P[j].z - P[j - 1].z;

	float a = d2_x * d2_x + d2_y * d2_y + d2_z * d2_z;
	float b = -2 * ((d1_x + pq_x)*d2_x + (d1_y + pq_y)*d2_y + (d1_z + pq_z)*d2_z);
	float c = d1_x*d1_x + d1_y*d1_y + d1_z*d1_z + 2 * (pq_x*d1_x + pq_y*d1_y + pq_z*d1_z) +
		pq_x*pq_x + pq_y * pq_y + pq_z*pq_z - _epsilon * _epsilon;
	float d = b*b - 4 * a*c;

	if (fabs(d) <= 1e-8) {
		float root = -b / (2 * a);
		if (root < 0 || root > 1)
			return false;
		else if (fabs(root - 1.0) > 1e-8)
			return false;
	}
	else {
		if (d < 0)
			return false;
		else {
			float root1 = (-b + sqrt(d)) / (2 * a);
			float root2 = (-b - sqrt(d)) / (2 * a);
			if (root1 > root2)
				std::swap(root1, root2);

			if (root1 > 1 && root2 > 1 || root1 < 0 && root2 < 0)
				return false;
			else if (root2 < 1)
				return false;
		}
	}

	B_F = F[1][0];
	if (B_F.first != 0)
		return false;
	else
		F[1][2] = B_F;

	std::pair<float, float> B_R_1, B_F_2;
	for (int k = 1; k < j - i; k++) {
		B_R_1 = F[k][2];
		B_F_2 = F[k + 1][0];
		if (B_F_2.second < B_R_1.first)
			return false;
		else
			F[k + 1][2] = make_pair(std::max(B_R_1.first, B_F_2.first), B_F_2.second);
	}

	return true;

}