#include "trainingdata.h"
#include "streamlinepool.h"
#include "entropycalculator.h"
#include "curvtorscalculator.h"
#include "tortuositycalculator.h"
#include "segmenthistcalculator.h"
#include "minvolenclosingellipsoid.h"
#include <fstream>
#include <iostream>
#include <cassert>
#include <cmath>
#include <numeric>

#include "eval.h"


Trainer::Trainer()
: _currentSid(-1), _accuracy(10), _lastSegPoint(-1), _segPointModel(nullptr),  _trainingFinished(false),
_numberOfNeighborhoods(4), _startPercentage(5)
{
	// load model from previous training	
	_segPointModel = svm_load_model("seg_points.model");
	cout << "segPoint model loaded successfully" << endl;
	int numPos = _segPointModel->nSV[0];
	int numNeg = _segPointModel->nSV[1];
}

Trainer::~Trainer()
{
	if (_segPointModel) {
		svm_save_model("C:/Users/lyifei/work/FlowVis/training/seg_points.model", _segPointModel);
		svm_free_and_destroy_model(&_segPointModel);
	}
}

void Trainer::genernateNHS(vector<int>& nhs, int sid, int center) {
	int numPts = _pStreamlines->streamlines[sid]->numPts;

	float percentage = (float)_startPercentage / 100.0;
	for (int i = 0; i < _numberOfNeighborhoods; i++) 
		nhs.push_back(int(numPts * (i+1) * percentage));


	for (auto &i : nhs) {
		if (center - i < 0)
			i = center;
		if (center + i > numPts - 1)
			i = numPts - 1 - center;
	}
}

// id is the index of simplification points
bool Trainer::compute_point_features(int sid, int id, vector<float>& feature)
{
	// compute entropy and tortuosity ratio using different neighborhood size
	vector<int> nhs;
	genernateNHS(nhs, sid, id);

	// velocity entropy
	feature.push_back(EntropyCalculator::instance().computeEntropyRatio(sid, id, nhs.front()));
	for (size_t i = 1; i < nhs.size(); i++) {
		if (nhs[i] == nhs[i - 1])
			feature.push_back(feature.back());
		else
			feature.push_back(EntropyCalculator::instance().computeEntropyRatio(sid, id, nhs[i]));
	}
	

	// tortuosity
	feature.push_back(TortuosityCalculator::instance().computeTortuosityRatio(sid, id, nhs.front()));
	for (size_t i = 1; i < nhs.size(); i++) {
		if (nhs[i] == nhs[i - 1])
			feature.push_back(feature.back());
		else
			feature.push_back(TortuosityCalculator::instance().computeTortuosityRatio(sid, id, nhs[i]));
	}
	
	// curvature and torsion histogram
	vector<float> left_hist, right_hist;
	SegmentHistogramCalculator::instance().computeHistogram(sid, id - nhs.front(), id, left_hist);
	SegmentHistogramCalculator::instance().computeHistogram(sid, id, id + nhs.front(), right_hist);
	feature.push_back(segment_dist(left_hist, right_hist));
	for (size_t i = 1; i < nhs.size(); i++) {
		if (nhs[i] == nhs[i - 1])
			feature.push_back(feature.back());
		else {
			left_hist.clear();
			right_hist.clear();
			SegmentHistogramCalculator::instance().computeHistogram(sid, id - nhs[i], id, left_hist);
			SegmentHistogramCalculator::instance().computeHistogram(sid, id, id + nhs[i], right_hist);
			feature.push_back(segment_dist(left_hist, right_hist));
		}
	}
	


	// min volume ratio	
	feature.push_back(MVEE::instance().computeBoundingEllipsoidVolRatio(sid, id, nhs.front()));
	for (size_t i = 1; i < nhs.size(); i++) {
		if (nhs[i] == nhs[i - 1])
			feature.push_back(feature.back());
		else
			feature.push_back(MVEE::instance().computeBoundingEllipsoidVolRatio(sid, id, nhs[i]));
	}
	

	return true;

}


void Trainer::add_seg_point_example(int id, bool isSegPoint)
{
	vector<float> features;

	if (isSegPoint) {
		features.push_back(1);
		_exampleRecords[_currentSid].seg_point_ids.push_back(id);
	}
	else {
		features.push_back(-1);
		_exampleRecords[_currentSid].non_seg_point_ids.push_back(id);
	}


	if( compute_point_features(_currentSid, id, features) )
		_segPointExamples.push_back(features);
}

// left/right is the number of points to the left/right of 'id' which are also considered as segmentation points
void Trainer::add_seg_point_range(int id, int left, int right)
{
	_segPointRange[id] = make_pair(left, right);
	for (int i = id-left; i <= id+right; i++) {
		if (i != id)
			add_seg_point_example(i, true);
	}
}


void Trainer::addNegativeSegPointTrainingData(const vector<int>& segpoints)
{
	Attribute<unsigned>* simplification_pts = dynamic_cast<Attribute<unsigned>*>(_pStreamlines->streamlines[_currentSid]->attributes["simplification_points"]);
	if (simplification_pts->size() <= 2)
		return;

	vector<unsigned> indices = simplification_pts->data();
	int numPts = _pStreamlines->streamlines[_currentSid]->numPts;

	int start, end;
	for (int i = 0; i < (int)segpoints.size() - 1; i++) {
		start = segpoints[i];
		end = segpoints[i + 1];
	
		auto it1 = lower_bound(indices.begin(), indices.end(), start);
		
		if (i == 0)
			it1++;
		auto it2 = lower_bound(indices.begin(), indices.end(), end);
		
		
		int n1, n2;
		n1 = n2 = 0;
	
		if (_segPointRange.find(start) != _segPointRange.end()) {
			n1 = _segPointRange[start].second + 1;
			start = *(it1 + _segPointRange[start].second);
		}
		
		if (_segPointRange.find(end) != _segPointRange.end()) {
			n2 = _segPointRange[end].first + 1;
			end = *(it2 - _segPointRange[end].first);
		}

		it1 = it1 + n1;
		while ((it1 != indices.end()) && (*it1 - start <= numPts * 0.05))
			it1++;

		if (it1 == indices.end())
			it1 = lower_bound(indices.begin(), indices.end(), start) + n1;

		it2 = it2 - n2;
		while ((it2 > it1) && (end - *it2 <= numPts * 0.05) )
			it2--;

		if (it2 == it1) {
			it2 = lower_bound(indices.begin(), indices.end(), end) - n2;
		}

		if (it1 == it2)
			return;

		if (*it2 == simplification_pts->back())
			--it2;
	}

}


void Trainer::train()
{
	int num_positive = count_if(_segPointExamples.begin(), _segPointExamples.end(), [](const vector<float>& e) { return e.front() == 1; });
	int num_negative = _segPointExamples.size() - num_positive;
	
	compute_segpoint_classifier();
	
	_trainingFinished = true;
}

void Trainer::reTrain()
{
	_segPointExamples.clear();
	for (auto& kv : _exampleRecords) {
		int sid = kv.first;
		for (auto id : kv.second.seg_point_ids) {
			vector<float> feature;
			feature.push_back(1);
			compute_point_features(sid, id, feature);
			_segPointExamples.push_back(feature);
		}

		for (auto id : kv.second.non_seg_point_ids) {
			vector<float> feature;
			feature.push_back(-1);
			compute_point_features(sid, id, feature);
			_segPointExamples.push_back(feature);
		}
	}

	train();
}




void Trainer::compute_segpoint_classifier()
{
	//// 1. use class-weighted SVM to handle unbalanced training examples
	//double best_auc1;
	//struct svm_model* best_model1 = rbf_SVM(_segPointExamples, true, best_auc1);
	//cout << "SegPoint classifier trained using class-weighted RBF: " << best_auc1 << endl;

	//2. GSVM-RU algorithm
	vector<vector<float>> aggregation;
	copy_if(_segPointExamples.begin(), _segPointExamples.end(), back_inserter(aggregation), [](const vector<float>& example){return example.front() == 1.0; });

	double best_auc2 = 0;
	struct svm_model* best_model2 = nullptr;
	vector<vector<float>> examples = _segPointExamples;
	
	int iter = 0;
	while (true) {
		iter++;
		struct svm_model* model = linear_SVM(examples, false);

		// remove old negative SVs
		aggregation.erase(find_if(aggregation.begin(), aggregation.end(), [](const vector<float>& e) { return e.front() == -1.0; }), aggregation.end());

		// add negative support vectors into aggregation
		unordered_set<int> indices_to_remove;
		for (int i = 0; i < model->l; i++) {
			if (examples[model->sv_indices[i] - 1].front() == -1.0) {
				aggregation.push_back(examples[model->sv_indices[i] - 1]);
				indices_to_remove.insert(model->sv_indices[i] - 1);
			}
		}

		// remove negative examples from training data
		vector<vector<float>> tmp;
		for (size_t i = 0; i < examples.size(); i++) {
			if (indices_to_remove.find(i) == indices_to_remove.end())
				tmp.push_back(examples[i]);
		}
		examples = tmp;


		svm_free_and_destroy_model(&model);

		// continue if classification performance improves
		double auc;
		model = rbf_SVM(aggregation, false, auc);
		if (auc <= best_auc2)
			break;
		else {
			best_auc2 = auc;
			if (best_model2)
				svm_free_and_destroy_model(&best_model2);
			best_model2 = model;
		}
	}
	std::cout << "SegPoint classifier trained using GSVM-RU: " << best_auc2 << " # iters: " << iter << endl;

	if (_segPointModel)
		svm_free_and_destroy_model(&_segPointModel);
	_segPointModel = best_model2;

}


struct svm_model* Trainer::linear_SVM(const vector<vector<float>>& examples, bool weighting)
{
	struct svm_problem problem;
	problem.l = examples.size();
	problem.y = new double[problem.l];
	problem.x = new svm_node*[problem.l];
	for (int i = 0; i < problem.l; i++)
		(problem.x)[i] = new svm_node[examples.front().size()];

	for (size_t i = 0; i < examples.size(); i++) {
		problem.y[i] = examples[i].front();

		size_t j = 0;
		for (; j < examples[i].size()-1; j++) {
			(problem.x)[i][j].index = j+1;
			(problem.x)[i][j].value = examples[i][j+1];
		}
		(problem.x)[i][j].index = -1;
	}

	vector<double> C{ 2e-10, 2e-9, 2e-8, 2e-7, 2e-6, 2e-5, 2e-4, 2e-3, 2e-2, 2e-1, 1, 2, 2e2};

	struct svm_parameter param;
	param.svm_type = C_SVC;
	param.kernel_type = LINEAR;
	param.eps = 1e-3;
	param.degree = 3;
	param.coef0 = 0;
	param.nu = 0.5;
	param.p = 0.1;
	param.gamma = 0;
	param.probability = 0;
	param.shrinking = 1;
	param.cache_size = 100;
	if (weighting) {
		param.nr_weight = 2;
		param.weight_label = new int[2];
		param.weight = new double[2];
		param.weight_label[0] = 1;
		param.weight_label[1] = -1;
	}
	else {
		param.nr_weight = 0;
		param.weight = 0;
		param.weight_label = 0;
	}

	int num_positive, num_negative;
	if (weighting) {
		num_positive = count_if(examples.begin(), examples.end(), [](const vector<float>& e) { return e.front() == 1; });
		num_negative = examples.size() - num_positive;
	}

	double best_auc = 0;
	double best_C;
	for (double c : C) {
		param.C = c;
		if (weighting) {
			param.weight[0] = (double)num_positive / c;
			param.weight[1] = (double)num_negative / c;
		}

		const char* msg = svm_check_parameter(&problem, &param);
		assert(msg == 0);
		double auc = binary_class_cross_validation(&problem, &param, 4);

		if (auc > best_auc) {
			best_auc = auc;
			best_C = c;
		}
	}

	param.C = best_C;
	struct svm_model *model = svm_train(&problem, &param);


	return model;
}


struct svm_model* Trainer::rbf_SVM(const vector<vector<float>>& examples, bool weighting, double& auc)
{
	struct svm_problem problem;
	problem.l = examples.size();
	problem.y = new double[problem.l];
	problem.x = new svm_node*[problem.l];
	for (int i = 0; i < problem.l; i++)
		(problem.x)[i] = new svm_node[examples.front().size()];

	int i = 0;
	for (auto& e : examples) {
		problem.y[i] = e.front();

		size_t j = 0;
		for (; j < e.size()-1; j++) {
			(problem.x)[i][j].index = j+1;
			(problem.x)[i][j].value = e[j+1];
		}
		(problem.x)[i][j].index = -1;
		i++;
	}

	int num_positive, num_negative;
	if (weighting) {
		num_positive = count_if(examples.begin(), examples.end(), [](const vector<float>& e) { return e.front() == 1; });
		num_negative = examples.size() - num_positive;
	}

	struct svm_parameter param;
	param.svm_type = C_SVC;
	param.kernel_type = RBF;
	param.eps = 1e-3;
	param.degree = 3;
	param.coef0 = 0;
	param.nu = 0.5;
	param.p = 0.1;
	param.probability = 0;
	param.shrinking = 1;
	param.cache_size = 100;
	if (weighting) {
		param.nr_weight = 2;
		param.weight_label = new int[2];
		param.weight = new double[2];
		param.weight_label[0] = 1;
		param.weight_label[1] = -1;
	}
	else {
		param.nr_weight = 0;
		param.weight = 0;
		param.weight_label = 0;
	}

	vector<double> C{ 2e-5, 2e-3, 2e-1, 1, 2e1, 2e3, 2e5, 2e7, 2e9, 2e11, 2e13 };
	vector<double> gamma{ 2e-15, 2e-13, 2e-11, 2e-9, 2e-7, 2e-5, 2e-3, 2e-1, 1, 2e1, 2e3 };

	double best_auc = 0;
	double best_C, best_gamma;
	for (auto c : C) {
		param.C = c;
		if (weighting) {
			param.weight[0] = (double)num_positive / c;
			param.weight[1] = (double)num_negative / c;
		}

		for (auto r : gamma) {
			param.gamma = r;

			const char* msg = svm_check_parameter(&problem, &param);
			assert(msg == 0);
			double cv = binary_class_cross_validation(&problem, &param, 4);
			if (cv > best_auc) {
				best_auc = cv;
				best_C = c;
				best_gamma = r;
			}
		}
	}

	param.C = best_C;
	param.gamma = best_gamma;
	auc = best_auc;

	struct svm_model* model = svm_train(&problem, &param);

	return model;
}

float Trainer::segment_dist(const vector<float>& hist1, const vector<float>& hist2) {
	
	assert(hist1.size() == hist2.size());

	float cdf1, cdf2, dist;
	cdf1 = cdf2 = dist = 0;
	for (size_t i = 0; i < hist1.size(); i++) {
		cdf1 += hist1[i];
		cdf2 += hist2[i];

		dist += fabs(cdf1 - cdf2);
	}

	return dist;
	
}

