#include "UnitSpherePartitioner.h"
#include <cmath>
#include <vector>
#include <algorithm>
#include <numeric>



#define PI 3.14159

UnitSpherePartitioner& UnitSpherePartitioner::instance()
{
	static UnitSpherePartitioner partitioner;
	return partitioner;
}

UnitSpherePartitioner::UnitSpherePartitioner()
{
	partition(100);
}

float UnitSpherePartitioner::sphericalCapArea(float theta)
{
	return 4.0f*PI*(sinf(theta / 2.0f) * sinf(theta / 2.0f));
}

float UnitSpherePartitioner::sphericalRadius(float area)
{
	return asinf(sqrt(area / (4.0f*PI))) * 2.0f; 
}

void UnitSpherePartitioner::partition(int numPartition)
{
	N = numPartition;

	// clean-up
	m.clear();
	theta.clear();
	alpha.clear();

	// determine the colatitudes of polar caps

	float vr = 4.0f * PI / N;  // area of each region

	theta_c = sphericalRadius(vr);

	// determine an ideal colloar angle
	float delta_I = sqrt(vr);

	// determine an ideal number of collars
	int n_I = int((PI - 2 * theta_c) / delta_I);

	// determine actual number of collars
	int n = std::max(1, int(floor(n_I + 0.5)));

	// create a list of the ideal number of regions in each collar
	delta_F = (PI - 2.0*theta_c) / n;

	std::vector<float> theta_F;
	for (int i = 1; i < n + 2; i++)
		theta_F.push_back(theta_c + (i - 1)*delta_F);

	std::vector<float> y;
	for (int i = 0; i < n; i++)
		y.push_back((sphericalCapArea(theta_F[i + 1]) - sphericalCapArea(theta_F[i])) / vr);

	// create a list of the actual number of regions in each collar
	std::vector<float> a;
	a.resize(n + 1);
	a[0] = 0;
	for (int i = 0; i < n; i++) {
		m.push_back(int(floor(y[i] + a[i] + 0.5)));
		float sum = 0;
		for (int j = 0; j < i + 1; j++)
			sum += y[j] - m[j];
		a.push_back(sum);
	}

	// create a list of colatitudes of each zone
	for (int i = 0; i < n + 1; i++) {
		int init = 0;
		float t = sphericalRadius((std::accumulate(m.begin(), m.begin() + i, init) + 1.0) * vr);
		theta.push_back(t);
	}

	// partition a collar
	for (size_t i = 0; i < m.size(); i++)
		alpha.push_back(PI*2.0f / m[i]);

}

int UnitSpherePartitioner::findRegionForVector(const float3 &v)
{
	float3 nv = normalize(v);
	float dp = dot(nv, make_float3(0, 1, 0));
	float t = acos(dp);
	dp = dot(nv, make_float3(0, 0, 1));
	float a = acos(dp);

	if (t < theta_c)
		return 0;
	else if (t >= theta.back())
		return N - 1;
	else {
		int i = 0;
		int j = theta.size() - 1;
		int c, k;
		while (true) {
			if (i + 1 == j) {
				c = i;
				break;
			}

			k = (i + j) / 2;
			if (theta[k] < t)
				i = k;
			else if (theta[k] > t)
				j = k;
			else {
				c = k;
				break;
			}
		}

		int init = 0;
		return std::accumulate(m.begin(), m.begin() + c, init) + int(a / alpha[c]);
	}
}
